/*-
 * Copyright (c) 1982, 1986, 1991, 1993
 *	The Regents of the University of California.  All rights reserved.
 * (c) UNIX System Laboratories, Inc.
 * All or some portions of this file are derived from material licensed
 * to the University of California by American Telephone and Telegraph
 * Co. or Unix System Laboratories, Inc. and are reproduced herein with
 * the permission of UNIX System Laboratories, Inc.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *	@(#)kern_subr.c	8.3 (Berkeley) 1/21/94
 */

#include <sys/types.h>
#include <sys/queue.h>
#include <stdio.h>
#include <stdlib.h>

#include <assert.h>

#include <netinet/in.h>
#include <arpa/inet.h>

#include <strings.h>


struct malloc_type;
void	hashdestroy(void *, struct malloc_type *, u_long);
void	*hashinit(int count, struct malloc_type *type, u_long *hashmask);
void	*hashinit_flags(int count, struct malloc_type *type,
    u_long *hashmask, int flags);
#define	HASH_NOWAIT	0x00000001
#define	HASH_WAITOK	0x00000002

void	*phashinit(int count, struct malloc_type *type, u_long *nentries);
#define powerof2(x)	((((x)-1)&(x))==0)
#define    powerof2rdown(x)    (powerof2(x) ? (x) : (0x1 << (fls((x) - 1) - 1)))
#define    powerof2rdfloor(x, y)    (MAX(powerof2rdown(x), (y))) 

#define kmalloc(size, type, wait)    my_malloc(size)
#define kfree(ptr, type)    free(ptr)

void *my_malloc(size_t size) {return malloc(size);}

#define KASSERT(a, b) assert(a)

#define INP_PCBHASH(faddr, lport, fport, mask) \
        (((faddr) ^ ((faddr) >> 16) ^ ntohs((lport) ^ (fport))) & (mask))


/*
 * General routine to allocate a hash table with control of memory flags.
 */
void *
hashinit_flags(int elements, struct malloc_type *type, u_long *hashmask,
    int flags)
{
	long hashsize;
	LIST_HEAD(generic, generic) *hashtbl;
	int i;

	KASSERT(elements > 0, ("%s: bad elements", __func__));
	/* Exactly one of HASH_WAITOK and HASH_NOWAIT must be set. */
	KASSERT((flags & HASH_WAITOK) ^ (flags & HASH_NOWAIT),
	    ("Bad flags (0x%x) passed to hashinit_flags", flags));

	for (hashsize = 1; hashsize <= elements; hashsize <<= 1)
		continue;
	hashsize >>= 1;

	if (flags & HASH_NOWAIT)
		hashtbl = kmalloc((u_long)hashsize * sizeof(*hashtbl),
		    type, M_NOWAIT);
	else
		hashtbl = kmalloc((u_long)hashsize * sizeof(*hashtbl),
		    type, M_WAITOK);

	if (hashtbl != NULL) {
		for (i = 0; i < hashsize; i++)
			LIST_INIT(&hashtbl[i]);
		*hashmask = hashsize - 1;
	}
	return (hashtbl);
}

/*
 * Allocate and initialize a hash table with default flag: may sleep.
 */
void *
hashinit(int elements, struct malloc_type *type, u_long *hashmask)
{

	return (hashinit_flags(elements, type, hashmask, HASH_WAITOK));
}

void
hashdestroy(void *vhashtbl, struct malloc_type *type, u_long hashmask)
{
	LIST_HEAD(generic, generic) *hashtbl, *hp;

	hashtbl = vhashtbl;
	for (hp = hashtbl; hp <= &hashtbl[hashmask]; hp++)
		KASSERT(LIST_EMPTY(hp), ("%s: hash not empty", __func__));
	kfree(hashtbl, type);
}

static const int primes[] = { 1, 13, 31, 61, 127, 251, 509, 761, 1021, 1531,
			2039, 2557, 3067, 3583, 4093, 4603, 5119, 5623, 6143,
			6653, 7159, 7673, 8191, 12281, 16381, 24571, 32749 };
#define NPRIMES (sizeof(primes) / sizeof(primes[0]))

/*
 * General routine to allocate a prime number sized hash table.
 */
void *
phashinit(int elements, struct malloc_type *type, u_long *nentries)
{
	long hashsize;
	LIST_HEAD(generic, generic) *hashtbl;
	int i;

	KASSERT(elements > 0, ("%s: bad elements", __func__));
	for (i = 1, hashsize = primes[1]; hashsize <= elements;) {
		i++;
		if (i == NPRIMES)
			break;
		hashsize = primes[i];
	}
	hashsize = primes[i - 1];
	hashtbl = kmalloc((u_long)hashsize * sizeof(*hashtbl), type, M_WAITOK);
	for (i = 0; i < hashsize; i++)
		LIST_INIT(&hashtbl[i]);
	*nentries = hashsize;
	return (hashtbl);
}


struct mypcb {
	in_addr_t s_addr;
	in_port_t lport;
	in_port_t fport;
	LIST_ENTRY(mypcb) inp_hash;
};

LIST_HEAD(myinpcbhead, mypcb);
struct myinpcbhead *head;

int
main(int argc, char**argv)
{

	int maxsockets, scale, iter;
	int i, j;
	int hashsize;
	int hash_array_size;
	struct myinpcbhead *hashtbl;
	u_long hashmask;
	struct mypcb *allpcbs;
	long long int wasted_buckets;

	/*printf("sizeof(s_addr) = %d\nsizeof(lport/fport) = %d\n",
	    (int)sizeof(in_addr_t), (int)sizeof(in_port_t));
*/

	if ( argc < 3) {
		fprintf(stderr, "usage: %s maxsockets scale test_iterations\n", argv[0]);
		exit(1);
	}
	maxsockets = atoi(argv[1]);
	scale = atoi(argv[2]);
	iter = atoi(argv[3]);

	hash_array_size = maxsockets / scale;
	hashsize = powerof2rdown(hash_array_size);
	hashtbl = hashinit(hashsize, NULL, &hashmask);

	allpcbs = malloc(sizeof(struct mypcb) * maxsockets);

	for (i = 0; i < maxsockets; i++) {
		struct mypcb *pcb;

		pcb = &allpcbs[i];
		pcb->s_addr = random();
		pcb->lport = 80;
		pcb->fport = random();
		head = &hashtbl[INP_PCBHASH(pcb->s_addr, pcb->lport, pcb->fport, hashmask)];
		LIST_INSERT_HEAD(head, pcb, inp_hash);
	}

	wasted_buckets = 0;
	for (i = 0; i < hashsize; i++) {
		if (LIST_EMPTY(&hashtbl[i]))
		    wasted_buckets++;
	}
	{
		/*
		 * (kgdb) p sizeof(struct socket)
		 * $1 = 680
		 */
		long long int freebsd_socket_size = 680;
		/* total memory we've allocated for sockets under FreeBSD */
		long long int total_socket_size = freebsd_socket_size * maxsockets;
		/* total memory for the hash head */
		long long int total_hash_head_size = hashsize * sizeof(struct myinpcbhead);
		long long int total_memory_used = total_socket_size + total_hash_head_size;
		/* % of the hash heads that are unused */
		float pct_hash_heads_wasted =  ((100.0 * (float)wasted_buckets) / (float)hashsize);
		/* memory wasted by unused hash heads */	
		long long hash_head_memory_wasted = wasted_buckets * sizeof(struct myinpcbhead);
		/* memory wasted as a percentage of the total memory used by sockets and hashheads */
		float pct_memory_wasted_total = 
		    (100.0 * (float)hash_head_memory_wasted) / (float)total_memory_used;


	fprintf(stderr, "sizeof(freebsd_socket) = %d, sizeof(hashhead) = %d\n",
	    (int)freebsd_socket_size, (int)sizeof(struct myinpcbhead));
	fprintf(stderr, "allocated %d hashheads for %d sockets.\n", hashsize, maxsockets);
	fprintf(stderr, "memory wastage: %%%02.2f (%lld bytes out of %lld)\n",
	    pct_memory_wasted_total, hash_head_memory_wasted, total_memory_used);
	fprintf(stderr, "hash wastage: %%%02.2f (%lld bytes) (%lld wasted buckets out of %lld entries)\n",
	    pct_hash_heads_wasted,
	    wasted_buckets * sizeof(struct myinpcbhead),
	    wasted_buckets, hashsize);
	}

	for (j = 0; j < iter; j++) {
		for (i = 0; i < maxsockets; i++) {
			struct mypcb *pcb, *foundpcb;

			pcb = &allpcbs[i];
			head = &hashtbl[INP_PCBHASH(pcb->s_addr, pcb->lport, pcb->fport, hashmask)];
			LIST_FOREACH(foundpcb, head, inp_hash) {
				if (!memcmp(foundpcb, pcb, sizeof(*pcb)))
					break;
			}
			if (foundpcb == NULL) {
				fprintf(stderr, "can't find self in hash\n");
				abort();
			}
		}
		fprintf(stderr, "%d\n", j);
	}
}
